local ADDON_NAME	= "TextureIt"

----------------------------------------------------------
-- Set SHOW_TOOLTIPS to false, if you don't want to
-- see them anymore
----------------------------------------------------------
local SHOW_TOOLTIPS = true
----------------------------------------------------------

----------------------------------------------------------
local VERSION_SV	= 0.1 	-- DO NOT CHANGE
----------------------------------------------------------

local USE_TEXTURE_FILE_DIMENSIONS = true
local DEBUG_MODE 	= false
local MAX_SEARCH_RESULTS = 250
local MAX_ENTRIES_PER_NODE = 100
local SearchCounter = 0
local VERSION_CODE	= 0.85

local TEXTUREIT_DEFAULT_SAVED_VARS = {
	["TEXTUREIT_HIDE_TEXTURE_BACKDROP"] = false,
	["TEXTURE_IT_USE_TEXTURE_FILE_DIEMSNIONS"] = false,
}

local colorRed 			= "|cFF0000" 	-- Red
local colorDrkOrange 	= "|cFFA500"	-- Dark Orange
local colorTeal			= "|c00FFFF"	-- Addon Name color
local colorYellow 		= "|cFFFF00" 	-- yellow 
local TextureIt = ZO_Object:New()

local function debugMsg(msg)
	if not DEBUG_MODE then return end
	
	d(msg)
end

local function SetTexture(editBox)
	TextureIt_Animate_Stop()
	local text = string.lower(editBox:GetText())
	if not (text and type(text)=="string" and text ~= "") then return end
	
	if string.match(text, ".dds$") then
		TEXTUREIT.texture:SetTexture(text)
		TextureIt_ReanchorTexture()
	else
		TextureIt_TableSearch(text)
	end
end

local function SetUpSearchBox()
	local parent = TEXTUREIT.textureWinTree
	local label		= parent:GetNamedChild("Label")
	
	local cSearchBox = WINDOW_MANAGER:CreateControl(parent:GetName() .. "SearchBox", parent, CT_EDITBOX)
	TEXTUREIT.searchBox = cSearchBox
	cSearchBox:ClearAnchors()
	cSearchBox:SetAnchor(TOPLEFT, label, BOTTOMLEFT, 0, 15)
	cSearchBox:SetAnchor(BOTTOMRIGHT, label, BOTTOMRIGHT, 0, 37)
	cSearchBox:SetMouseEnabled(true)
	
	cSearchBox.bg = WINDOW_MANAGER:CreateControlFromVirtual(cSearchBox:GetName().."background", cSearchBox, "ZO_EditBackdrop")
	cSearchBox.bg:ClearAnchors()
	cSearchBox.bg:SetAnchorFill()
	cSearchBox.bg:SetDrawTier(1)
	
	cSearchBox.editbox = WINDOW_MANAGER:CreateControlFromVirtual(cSearchBox:GetName().."EditBox", cSearchBox, "ZO_DefaultEditForBackdrop")
	cSearchBox.editbox:ClearAnchors()
	cSearchBox.editbox:SetAnchorFill()
	cSearchBox.editbox:SetText(GetString(TEXTUREIT_PATH_WILL_BE_PLACED_HERE))
	
	cSearchBox.editbox:SetHandler("OnMouseDoubleClick", function() 
		cSearchBox.editbox:TakeFocus() 
		cSearchBox.editbox:SelectAll() 
	end)
	cSearchBox.editbox:SetHandler("OnEnter", SetTexture)
	
	cSearchBox.editbox.data = { tooltipText = GetString(TEXTUREIT_PATH_WILL_BE_PLACED_HERE_TT) }
	local tooltip = cSearchBox.editbox.data.tooltipText
	
    cSearchBox.editbox:SetHandler("OnMouseEnter", function()
		if not SHOW_TOOLTIPS then return end
		InitializeTooltip(InformationTooltip, cSearchBox, LEFT, 0, -2, RIGHT)
		SetTooltipText(InformationTooltip, tooltip)
	end)
	
    cSearchBox.editbox:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)
    cSearchBox.editbox:SetMaxInputChars(80)
	
	cSearchBox.label = WINDOW_MANAGER:CreateControl(cSearchBox:GetName() .. "Label", cSearchBox, CT_LABEL)
	cSearchBox.label:SetDimensions(200, 30)
	cSearchBox.label:ClearAnchors()
	cSearchBox.label:SetAnchor(TOPLEFT, cSearchBox, BOTTOMLEFT, 0, 5)
	cSearchBox.label:SetText("Not Found")
	cSearchBox.label:SetFont("ZoFontWinH4")
	cSearchBox.label:SetColor(1, 0, 0)
	cSearchBox.label:SetHidden(true)
end


function TextureIt:New()
    local obj 				= ZO_Object.New(self)
	self.textureWinTree 	= TextureItWinTree
	self.scrollChild 		= TextureItWinTreeTreeScrollChild
	self.editBox			= TextureItWinTreeEditBox
	
	self.textureWin			= TextureItWinTexture
	self.texture			= TextureItWinTextureTexture
	self.actualDimLabel		= TextureItWinTextureActualDimLabel
	self.curDimLabel		= TextureItWinTextureCurDimLabel
	
	self.animationWin		= AnimatePopUp
	return obj
end

-- Initialize everything else
function TextureIt:Initialize()
	self.sv = ZO_SavedVars:New("TextureItSavedVars", VERSION_SV, nil, TEXTUREIT_DEFAULT_SAVED_VARS)
	
	ZO_CreateStringId("SI_BINDING_NAME_TEXTUREIT_TOGGLE_MAIN_WINDOW", colorDrkOrange.."Toggles the entire TextureIt window.")
	
	TextureIt_ReanchorTexture()			-- Restore Texture size setting
	SetUpSearchBox()
end


function TextureIt:CreateTree()
    self.navigationTree = ZO_Tree:New(self.scrollChild, 25, 0, 1000)
	
    local openTexture 		= "EsoUI/Art/Buttons/tree_open_up.dds"
    local closedTexture 	= "EsoUI/Art/Buttons/tree_closed_up.dds"
    local overOpenTexture 	= "EsoUI/Art/Buttons/tree_open_over.dds"
    local overClosedTexture = "EsoUI/Art/Buttons/tree_closed_over.dds"
	
    local function TreeHeaderSetup(node, control, data, open)
        control.text:SetModifyTextType(MODIFY_TEXT_TYPE_UPPERCASE)
		control.text:SetFont("ZoFontWinH4")
        control.text:SetText(data.text)
        control.icon:SetTexture(open and openTexture or closedTexture)
        control.iconHighlight:SetTexture(open and overOpenTexture or overClosedTexture)
   
        control.text:SetSelected(open)
    end
	
    self.navigationTree:AddTemplate("TextureItHeader", TreeHeaderSetup, nil, nil, nil, 0)
	
    local function TreeEntrySetup(node, control, data, open)
		control:SetFont("ZoFontWinH4")
        control:SetText(data.text)
        control:SetSelected(false)
    end

    local function TreeEntryOnSelected(control, data, selected, reselectingDuringRebuild)
		TextureIt_Animate_Stop()
        control:SetSelected(selected)

        local texture = GetControl(control, "Icon")
        texture.selected = selected
        if selected then
			TEXTUREIT.texture:SetTexture(data.texturePath)
            texture:SetTexture("EsoUI/Art/Journal/journal_Quest_Selected.dds")
            texture:SetAlpha(1.00)
            texture:SetHidden(false)
			
			self.searchBox.editbox:SetText(data.texturePath)
			self.searchBox.editbox:TakeFocus()
			self.searchBox.editbox:SelectAll()
			-- The texture gets changed to slow for my function to get the
			-- correct file dimensions, so must delay a call to it
			zo_callLater(function() 
				TextureIt_ReanchorTexture()
			end, 30)
        else
            texture:SetHidden(true)
			TEXTUREIT.texture:SetTexture(nil)
        end
    end
    local function TreeEntryEquality(left, right)
        return left.name == right.name
    end
    self.navigationTree:AddTemplate("TextureItNavigationEntry", TreeEntrySetup, TreeEntryOnSelected, TreeEntryEquality)
	
	-- true means only one can be open at a time AND you can't click to close it. 
	-- You have to open another to get it to close
    --self.navigationTree:SetExclusive(true)
    self.navigationTree:SetExclusive(false)
    self.navigationTree:SetOpenAnimation("ZO_TreeOpenAnimation")
end
 
-- Used to copy the textures into a numerical table & sort it
local function SortTable(tTable)
	local sortedTable = {}
	
	for key, value in pairs(tTable) do
		table.insert(sortedTable, {key, value})
	end
	
	table.sort(sortedTable, function(val1, val2) return val1[1] < val2[1] end)
	
	return sortedTable
end

local function CountChildren(textureTable)
	local startTime =  GetGameTimeMilliseconds() 
	local totalCount = 0
	
	for k,v in pairs(textureTable) do
		if type(v) == "table" then
			totalCount = totalCount + CountChildren(v)
		else
			totalCount = totalCount + 1
		end
	end
	return totalCount
end

function TextureIt:AddTableNodes(textureTable, parentNode)
	local startTime =  GetGameTimeMilliseconds() 
	local sortedTable = SortTable(textureTable)
	-- Count of textures for THIS table were on
	local totalTextureCount = 0
	local localCount = 0
	
	for tableIndex=1, #sortedTable do
		--if totalTextureCount > 1500 then return localCount end
	
	   if type(sortedTable[tableIndex][2]) == "table" then
		   local childCount = CountChildren(sortedTable[tableIndex][2])
		   local childCountColor = "|cFFFFFF" -- white
			if childCount > 100 and childCount <= 500 then
			   childCountColor = colorYellow -- yellow
			elseif childCount > 500 then
			   childCountColor = colorRed -- red
			end
			local nextParentNode = self.navigationTree:AddNode("TextureItHeader", {["text"]=sortedTable[tableIndex][1].." (".. childCountColor ..childCount.."|r)", ["nodeTable"]=sortedTable[tableIndex][2]}, parentNode, nil)

		else
			local childNode = self.navigationTree:AddNode("TextureItNavigationEntry", {["text"]=sortedTable[tableIndex][1], ["texturePath"]=sortedTable[tableIndex][2]}, parentNode, SOUNDS.SKILL_LINE_SELECT)
			-- One texture found, add it to the counts
			localCount = localCount + 1
			totalTextureCount = totalTextureCount + 1
	   end
	end
	local endTime =  GetGameTimeMilliseconds() 
	debugMsg("Done adding Nodes, Time Elapsed: "..(endTime - startTime).." ms")
	debugMsg("Added "..totalTextureCount.." items.")
	return localCount
end

function TextureIt:PopulateTree()
	local startTime =  GetGameTimeMilliseconds() 
	
    TEXTUREIT.navigationTree:Reset()
	local textures = TEXTUREIT_TEXTURES
	
	self:AddTableNodes(textures, nil)
	
    self.navigationTree:Commit()
	
	local endTime =  GetGameTimeMilliseconds() 
	debugMsg("Done adding Nodes, Time Elapsed: "..(endTime - startTime).." ms")
	
	-- Update the windows heading label with the texture count:
	local headingLabel = self.textureWinTree:GetNamedChild("Label")
	local totalTextureCount = CountChildren(TEXTUREIT_TEXTURES)
	headingLabel:SetText(colorRed.."Circonians |r"..colorTeal.."TextureIt:|r"..colorYellow.." ("..totalTextureCount.." Textures)|r")
	--headingLabel:SetText(colorRed.."Circonians |r"..colorTeal.."TextureIt")
end
 
 
local function GetSearchResults(sourceTable, destTable, text)
	local savedTextures = sourceTable
	
	for key,value in pairs(savedTextures) do
		if SearchCounter >= MAX_SEARCH_RESULTS then
			return
		end
	   if type(value) == "table" then
			GetSearchResults(value, destTable, text)
			
		elseif type(value) == "string" then
			if string.lower(key):find(text) then
				table.insert(destTable, {["text"] = key, ["texturePath"] = value})
				SearchCounter = SearchCounter + 1
			end
	   end
	end
end

function TextureIt_TableSearch(text)
	local startTime =  GetGameTimeMilliseconds() 
	local sourceTable = TEXTUREIT_TEXTURES
	SearchCounter = 0
	local searchResults = {}
	
	GetSearchResults(sourceTable, searchResults, text)
	
	local function tableSort(tableA, tableB)
        return tableA.text < tableB.text
    end
	
	table.sort(searchResults, tableSort)

    TEXTUREIT.navigationTree:Reset()
	
	for key,value in ipairs(searchResults) do
		local childNode = TEXTUREIT.navigationTree:AddNode("TextureItNavigationEntry", {["text"]=value.text, ["texturePath"]=value.texturePath}, nil, SOUNDS.SKILL_LINE_SELECT)
	end
    TEXTUREIT.navigationTree:Commit()
	
	local endTime =  GetGameTimeMilliseconds()
	debugMsg("Search Done, Time Elapsed: "..(endTime - startTime).." ms")
	debugMsg("Found "..SearchCounter.." textures.")
end

---------------------------------------------------------
--  XML Functions --
---------------------------------------------------------
function TextureIt_Toggle_Animation_Window()
	TEXTUREIT.animationWin:SetHidden(not TEXTUREIT.animationWin:IsHidden())
end

function TextureIt_Animate_Stop()
	local animation = TEXTUREIT.texture.animation
	local timeline 	= TEXTUREIT.texture.timeline
	
	if not timeline or not animation then return end

	timeline:Stop() 		
	animation:SetImageData(1, 1)
	AnimatePopUpAnimateBtn:SetText("Animate")
end

function TextureIt_Animate_Texture(self)
	local numCellWide = self:GetParent():GetNamedChild("CellsWideEditBox"):GetText()
	local numCellHigh = self:GetParent():GetNamedChild("CellsHighEditBox"):GetText()
	local frameRate = self:GetParent():GetNamedChild("FrameRateEditBox"):GetText()
	
	numCellWide = tonumber(numCellWide)
	numCellHigh = tonumber(numCellHigh)
	frameRate = tonumber(frameRate)
	
	if type(numCellWide) ~= "number" or type(numCellHigh) ~= "number" 
	or type(frameRate) ~= "number" then return end
	
	local animation = TEXTUREIT.texture.animation
	local timeline 	= TEXTUREIT.texture.timeline
	
	if not animation then
		animation, timeline = CreateSimpleAnimation(ANIMATION_TEXTURE, TEXTUREIT.texture)
		TEXTUREIT.texture.animation = animation
		TEXTUREIT.texture.timeline = timeline
	end
	
	if timeline:IsPlaying() then 
		TextureIt_Animate_Stop()
	else
		animation:SetImageData(numCellWide, numCellHigh)
		animation:SetFramerate(frameRate)
		timeline:SetPlaybackType(ANIMATION_PLAYBACK_LOOP, LOOP_INDEFINITELY)
		timeline:PlayFromStart()
		
		AnimatePopUpAnimateBtn:SetText("Stop")
	end
end

function TextureIt_AnimationBtn_OnMouseEnter(self)
	if not SHOW_TOOLTIPS then return end
	
	local text = "Animate Texture"
	InitializeTooltip(InformationTooltip, self, LEFT, 25, 0, RIGHT)
	SetTooltipText(InformationTooltip, text)
end

function TextureIt_AnimationBtn_OnMouseExit(self)
    ClearTooltip(InformationTooltip)
end

function TextureIt_ToggleMainWindow()
	TEXTUREIT.textureWinTree:SetHidden(not TEXTUREIT.textureWinTree:IsHidden())
	TEXTUREIT.textureWin:SetHidden(not TEXTUREIT.textureWin:IsHidden())
end

function TextureIt_ToggleBackdrop()
	local backdrop = TEXTUREIT.textureWin:GetNamedChild("TextureBackdrop")
	
	backdrop:SetHidden(not backdrop:IsHidden())
end

function TextureIt_ReanchorTexture()
	local texture = TEXTUREIT.texture
	local textureWin = TEXTUREIT.textureWin
	
	if USE_TEXTURE_FILE_DIMENSIONS then
		local actWidth, actHeight = texture:GetTextureFileDimensions()
		local winWidth, winHeight = TEXTUREIT.textureWin:GetDimensions()
		local screenWidth, screenHeight = GuiRoot:GetDimensions()
		local textureBg 	= textureWin:GetNamedChild("TextureBackdrop")
		textureWin:SetDimensionConstraints(600, 400, 0, 0) 
		
		-- Some textures are extremely large. By reanchoring this ensures the position of the window
		-- and size, while keeping the labels at the top on the screen.
		textureWin:ClearAnchors()
		textureWin:SetAnchor(TOPLEFT, TEXTUREIT.textureWinTree, TOPRIGHT, 0, 0)
		-- Height: +50 for the border + 50 for offsetY, spacing for labels at the top
		textureWin:SetDimensions(actWidth+50, actHeight+100)
		
		texture:ClearAnchors()
		texture:SetAnchor(CENTER, textureBg, CENTER, 0, 0)
		texture:SetDimensions(actWidth, actHeight)
		TEXTUREIT.curDimLabel:SetText(GetString(TEXTUREIT_ACTUAL_DIMENSIONS)..zo_round(actWidth).." x "..zo_round(actHeight))
		TEXTUREIT.actualDimLabel:SetText(GetString(TEXTUREIT_CURRENT_DIMENSIONS)..zo_round(actWidth).." x "..zo_round(actHeight))
		
	else
		textureWin:SetDimensionConstraints(62, 112, 0, 0) 
		texture:ClearAnchors()
		texture:SetAnchor(TOPLEFT, textureBg, TOPLEFT, 25, 75)
		texture:SetAnchor(BOTTOMRIGHT, textureBg, BOTTOMRIGHT, -25, -25)
		local curWidth, curHeight = texture:GetDimensions()
		local actWidth, actHeight = texture:GetTextureFileDimensions()
		TEXTUREIT.actualDimLabel:SetText("Actual Texture Dimensions: "..zo_round(actWidth).." x "..zo_round(actHeight))
		TEXTUREIT.curDimLabel:SetText("Current Texture Dimensions: "..zo_round(curWidth).." x "..zo_round(curHeight))
	end
end

function TextureIt_Toggle_Actual_Size()
	USE_TEXTURE_FILE_DIMENSIONS = not USE_TEXTURE_FILE_DIMENSIONS
	TextureIt_ReanchorTexture()
end

 
function TextureIt_On_ResizeStart(self)
	EVENT_MANAGER:RegisterForUpdate("TextureIt_TextureResize", 50, function() 
		local curWidth, curHeight = TEXTUREIT.texture:GetDimensions()
		
		TEXTUREIT.curDimLabel:SetText("Current Texture Dimensions: "..zo_round(curWidth).." x "..zo_round(curHeight))
	end)
 end
function TextureIt_On_ResizeStop(self)
	EVENT_MANAGER:UnregisterForUpdate("TextureIt_TextureResize")
 end
 
function TextureItNavigationEntry_OnMouseUp(label, button, upInside)
	if not upInside then return end
	if button == 1 then
		if(upInside and label.node.tree.enabled) then
			-- Play the selected sound if not already selected
			if(not label.node.selected and label.node.selectSound) then
				PlaySound(label.node.selectSound)
			end
			--ZO_TreeEntry_OnMouseUp(label, upInside)
			label.node:GetTree():ToggleNode(label.node)
			label.node:GetTree():SelectNode(label.node)
		end
    end
end
 
-- This doesn't work anymore since the changes I made
-- But doesn't hurt anything, I need to change this.TextureIt_Toggle_Actual_Size
function TextureItNavigationEntry_OnMouseEnter(texture)
    if texture.tooltipText and texture.tooltipText ~= "" then
        InitializeTooltip(InformationTooltip, texture, BOTTOM, 0, 0, TOP)
        SetTooltipText(InformationTooltip, texture.tooltipText)
    end
end
 
function TextureItNavigationEntry_OnMouseExit()
    ClearTooltip(InformationTooltip)
end
 
local function Header_OnMouseEnter(control)
	ZO_SelectableLabel_OnMouseEnter(control.text)
	control.iconHighlight:SetHidden(false)
end

local function Header_OnMouseExit(control)
	ZO_SelectableLabel_OnMouseExit(control.text)
	control.iconHighlight:SetHidden(true)
end

local function Header_OnMouseUp(control, upInside)
	if not upInside then return end
		local node = control.node
	if not node.open and not node.children then
		TEXTUREIT:AddTableNodes(node.data.nodeTable, node)
	end
	ZO_TreeHeader_OnMouseUp(control, upInside)
end

function TextureItHeader_OnInitialized(self)
	self.icon = self:GetNamedChild("Icon")
	self.iconHighlight = self.icon:GetNamedChild("Highlight")
	self.text = self:GetNamedChild("Text")

	self.OnMouseEnter = Header_OnMouseEnter
	self.OnMouseExit = Header_OnMouseExit
	self.OnMouseUp = Header_OnMouseUp
end
 

-------------------------------------------------------------------
--  OnAddOnLoaded  --
-------------------------------------------------------------------
local function OnAddOnLoaded(event, addonName)
	if addonName == ADDON_NAME then
		TEXTUREIT = TextureIt:New()
		TEXTUREIT:Initialize()
		TEXTUREIT:CreateTree()
		TEXTUREIT:PopulateTree()	
	end
end

---------------------------------------------------------------------
--  Register Events --
---------------------------------------------------------------------
EVENT_MANAGER:RegisterForEvent(ADDON_NAME, EVENT_ADD_ON_LOADED, OnAddOnLoaded)


---------------------------------------------------------------------
--  Slash Commands --
---------------------------------------------------------------------
SLASH_COMMANDS["/textureit"] 		= TextureIt_ToggleMainWindow
SLASH_COMMANDS["/texit"] 			= TextureIt_ToggleMainWindow
